/*
 *
 *  BlueZ - Bluetooth protocol stack for Linux
 *
 *  Copyright (C) 2005-2007  Marcel Holtmann <marcel@holtmann.org>
 *  Copyright (C) 2006-2007  Bastien Nocera <hadess@hadess.net>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dbus/dbus-glib.h>

#include <glib/gi18n.h>

#include <gtk/gtk.h>

#include "general.h"

static DBusGConnection *connection = NULL;

static GtkWidget *button_properties;

static GtkListStore *service_store;

static void startstop_notify(DBusGProxy *object,
				DBusGProxyCall *call, void *user_data)
{
	gchar *path = user_data;
	GtkTreeIter iter;
	gboolean active;

	dbus_g_proxy_end_call(object, call, NULL, G_TYPE_INVALID);

	if (gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(service_store),
							&iter, path) == FALSE)
		return;

	dbus_g_proxy_call(object, "IsRunning", NULL, G_TYPE_INVALID,
				G_TYPE_BOOLEAN, &active, G_TYPE_INVALID);

	gtk_list_store_set(service_store, &iter, 2, active, 3, FALSE, -1);
}

static void startstop_callback(GtkCellRendererToggle *renderer,
					gchar *path, gpointer user_data)
{
	DBusGProxy *object;
	GtkTreeIter iter;
	const char *service_path;
	gboolean active;

	if (connection == NULL)
		return;

	if (gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(service_store),
							&iter, path) == FALSE)
		return;

	gtk_tree_model_get(GTK_TREE_MODEL(service_store),
				&iter, 0, &service_path, 2, &active, -1);

	object = dbus_g_proxy_new_for_name(connection, "org.bluez",
					service_path, "org.bluez.Service");

	gtk_list_store_set(service_store, &iter, 3, TRUE, -1);

	if (active == FALSE) {
		dbus_g_proxy_begin_call(object, "Start", startstop_notify,
				g_strdup(path), g_free, G_TYPE_INVALID);
		active = TRUE;
	} else {
		dbus_g_proxy_begin_call(object, "Stop", startstop_notify,
				g_strdup(path), g_free, G_TYPE_INVALID);
		active = FALSE;
	}
}

static void status_to_toggle(GtkTreeViewColumn *column, GtkCellRenderer *cell,
			GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gboolean active, progress;

	gtk_tree_model_get(model, iter, 2, &active, 3, &progress, -1);

	g_object_set(cell, "active", active, NULL);
	g_object_set(cell, "sensitive", progress == TRUE ? FALSE : TRUE, NULL);
}

static void status_to_text(GtkTreeViewColumn *column, GtkCellRenderer *cell,
			GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gboolean active, progress;
	gchar *str;

	gtk_tree_model_get(model, iter, 2, &active, 3, &progress, -1);

	if (active == TRUE) {
		if (progress == TRUE)
			str = "Stopping";
		else
			str = "Running";
	} else {
		if (progress == TRUE)
			str = "Starting";
		else
			str = "Stopped";
	}
 
	g_object_set(cell, "text", str, NULL);
}

static void show_properties(const gchar *path, const gchar *description)
{
	GtkWidget *dialog;
	GtkWidget *button;
	GtkWidget *vbox;
	GtkWidget *label;
	gint response;

	dialog = gtk_dialog_new();

	gtk_window_set_title(GTK_WINDOW(dialog), _("Service properties"));

	gtk_window_set_icon_name(GTK_WINDOW(dialog), "stock_bluetooth");

	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER_ON_PARENT);

	gtk_window_set_default_size(GTK_WINDOW(dialog), 240, -1);

	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);

	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
					GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE);

	gtk_widget_grab_default(button);

	vbox = gtk_vbox_new(FALSE, 6);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), vbox);

	label = create_label(_("Service path"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	label = gtk_label_new(path);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);

	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	label = create_label(_("Description"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	label = gtk_label_new(description);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);

	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	gtk_widget_show_all(GTK_WIDGET(dialog));

	response = gtk_dialog_run(GTK_DIALOG(dialog));

	gtk_widget_destroy(GTK_WIDGET(dialog));
}

static void properties_callback(GtkWidget *button, gpointer user_data)
{
	GtkTreeSelection *selection = user_data;
	GtkTreeIter iter;
	GtkTreeModel *model;
	gchar *path, *description;

	if (gtk_tree_selection_get_selected(selection, &model, &iter) == TRUE) {
		gtk_tree_model_get(model, &iter, 0, &path, 4, &description, -1);

		show_properties(path, description);

		g_free(path);
	}
}

static void select_callback(GtkTreeSelection *selection, gpointer user_data)
{
	gboolean sensitive;

	sensitive = gtk_tree_selection_get_selected(selection, NULL, NULL);

	gtk_widget_set_sensitive(GTK_WIDGET(button_properties), sensitive);
}

GtkWidget *create_service(void)
{
	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *scrolled;
	GtkWidget *tree;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;
	GtkWidget *buttonbox;

	if (connection == NULL)
		return NULL;

	vbox = gtk_vbox_new(FALSE, 6);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	label = create_label(_("Available services"));
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled),
							GTK_SHADOW_OUT);
	gtk_container_add(GTK_CONTAINER(vbox), scrolled);

	tree = gtk_tree_view_new();
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree), TRUE);
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tree), TRUE);
	gtk_tree_view_set_model(GTK_TREE_VIEW(tree),
					GTK_TREE_MODEL(service_store));

	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), 0,
					"Name", gtk_cell_renderer_text_new(),
							"text", 1, NULL);
	column = gtk_tree_view_get_column(GTK_TREE_VIEW(tree), 0);
	gtk_tree_view_column_set_expand(GTK_TREE_VIEW_COLUMN(column), TRUE);

	column = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title(column, "Status");
	gtk_tree_view_column_set_spacing(column, 2);

	renderer = gtk_cell_renderer_toggle_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_cell_data_func(column, renderer,
						status_to_toggle, NULL, NULL);

	g_signal_connect(G_OBJECT(renderer), "toggled",
				G_CALLBACK(startstop_callback), NULL);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_set_cell_data_func(column, renderer,
						status_to_text, NULL, NULL);

	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

	gtk_tree_view_column_set_min_width(GTK_TREE_VIEW_COLUMN(column), 100);
	gtk_tree_view_column_set_max_width(GTK_TREE_VIEW_COLUMN(column), 100);

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
	g_signal_connect(G_OBJECT(selection), "changed",
				G_CALLBACK(select_callback), NULL);

	gtk_container_add(GTK_CONTAINER(scrolled), tree);

	buttonbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(buttonbox),
						GTK_BUTTONBOX_START);
	gtk_box_pack_start(GTK_BOX(vbox), buttonbox, FALSE, FALSE, 0);

	button_properties = gtk_button_new_from_stock(GTK_STOCK_PROPERTIES);
	gtk_widget_set_sensitive(GTK_WIDGET(button_properties), FALSE);
	gtk_container_add(GTK_CONTAINER(buttonbox), button_properties);
	g_signal_connect(G_OBJECT(button_properties), "clicked",
				G_CALLBACK(properties_callback), selection);

	return vbox;
}

static GList *service_list = NULL;

struct service_data {
	char *path;
	int present;
	GtkTreeIter iter;
};

static void service_free(gpointer data, gpointer user_data)
{
	struct service_data *service = data;

	service_list = g_list_remove(service_list, service);

	g_free(service->path);
	g_free(service);
}

static void service_disable(gpointer data, gpointer user_data)
{
	struct service_data *service = data;

	if (service->present) {
		service->present = 0;
		gtk_list_store_remove(service_store, &service->iter);
	}
}

static gint service_compare(gconstpointer a, gconstpointer b)
{
	const struct service_data *service = a;
	const char *path = b;

	return g_ascii_strcasecmp(service->path, path);
}

static void service_count(gpointer data, gpointer user_data)
{
	struct service_data *service = data;
	int *count = user_data;

	if (service->present)
		(*count)++;
}

static int present_services(void)
{
	int count = 0;

	g_list_foreach(service_list, service_count, &count);

	return count;
}

static void service_started(DBusGProxy *object, gpointer user_data)
{
	struct service_data *service = user_data;

	gtk_list_store_set(service_store, &service->iter, 2, TRUE, -1);
}

static void service_stopped(DBusGProxy *object, gpointer user_data)
{
	struct service_data *service = user_data;

	gtk_list_store_set(service_store, &service->iter, 2, FALSE, -1);
}

static void add_service(const char *path)
{
	DBusGProxy *object;
	GList *list;
	struct service_data *service;
	const char *name = NULL;
	const char *description = NULL;
	gboolean running;

	list = g_list_find_custom(service_list, path, service_compare);
	if (list && list->data) {
		service = list->data;
		goto done;
	}

	service = g_try_malloc0(sizeof(*service));
	if (!service)
		return;

	service->path = g_strdup(path);

	service_list = g_list_append(service_list, service);

done:
	service->present = 1;

	if (connection == NULL)
		return;

	object = dbus_g_proxy_new_for_name(connection, "org.bluez",
						path, "org.bluez.Service");

	dbus_g_proxy_add_signal(object, "Started", G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "Started",
				G_CALLBACK(service_started), service, NULL);

	dbus_g_proxy_add_signal(object, "Stopped", G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "Stopped",
				G_CALLBACK(service_stopped), service, NULL);

	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
				G_TYPE_STRING, &name, G_TYPE_INVALID);

	dbus_g_proxy_call(object, "GetDescription", NULL, G_TYPE_INVALID,
				G_TYPE_STRING, &description, G_TYPE_INVALID);

	dbus_g_proxy_call(object, "IsRunning", NULL, G_TYPE_INVALID,
				G_TYPE_BOOLEAN, &running, G_TYPE_INVALID);

	gtk_list_store_insert_with_values(service_store,
				&service->iter, -1, 0, path,
				1, name ? name : path, 2, running,
				4, description ? description : name, -1);
}

static void service_added(DBusGProxy *object,
				const char *path, gpointer user_data)
{
	add_service(path);
}

static void service_removed(DBusGProxy *object,
				const char *path, gpointer user_data)
{
	GList *list;

	list = g_list_find_custom(service_list, path, service_compare);
	if (list && list->data) {
		struct service_data *service = list->data;

		service->present = 0;

		gtk_list_store_remove(service_store, &service->iter);

		if (present_services() > 0)
			return;

		gtk_widget_set_sensitive(GTK_WIDGET(button_properties), FALSE);
	}
}

void setup_service(DBusGConnection *conn)
{
	DBusGProxy *proxy;
	GError *error = NULL;
	const gchar **array = NULL;

	service_store = gtk_list_store_new(5, G_TYPE_STRING, G_TYPE_STRING,
				G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_STRING);
	if (!service_store)
		return;

	proxy = dbus_g_proxy_new_for_name(conn, "org.bluez",
					"/org/bluez", "org.bluez.Manager");
	if (!proxy)
		return;

#if 0
	connection = dbus_g_connection_ref(conn);
#endif

	dbus_g_proxy_add_signal(proxy, "ServiceAdded",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(proxy, "ServiceAdded",
				G_CALLBACK(service_added), conn, NULL);

	dbus_g_proxy_add_signal(proxy, "ServiceRemoved",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(proxy, "ServiceRemoved",
				G_CALLBACK(service_removed), conn, NULL);

	dbus_g_proxy_call(proxy, "ListServices", &error,
			G_TYPE_INVALID, G_TYPE_STRV, &array, G_TYPE_INVALID);

	if (error == NULL) {
		while (*array) {
			add_service(*array);
			array++;
		}
	} else
		g_error_free(error);
}

void cleanup_service(void)
{
	g_list_foreach(service_list, service_free, NULL);

	if (connection != NULL)
		dbus_g_connection_unref(connection);
}

void disable_service(void)
{
	g_list_foreach(service_list, service_disable, NULL);

	if (connection != NULL)
		gtk_widget_set_sensitive(GTK_WIDGET(button_properties), FALSE);
}
